from typing import List, Dict, Optional
from datetime import datetime
import warnings
import os
from functools import wraps

from notion_client import Client
from notion_client.helpers import get_id

from notion_df.values import PageProperties, PageProperty
from notion_df.configs import DatabaseSchema

API_KEY = None
NOT_REVERSE_DATAFRAME = -1
# whether to reverse the dataframe when performing uploading.
# for some reason, notion will reverse the order of dataframe
# when uploading.
# -1 for reversing, for not reversing
NOTION_DEFAULT_PAGE_SIZE = 100
NOTION_MAX_PAGE_SIZE = 100


def config(api_key: str):
    global API_KEY
    API_KEY = api_key


def _load_api_key(api_key: str) -> str:
    if api_key is not None:
        return api_key
    elif API_KEY is not None:
        return API_KEY
    elif os.environ.get("NOTION_API_KEY") is not None:
        return os.environ.get("NOTION_API_KEY")
    else:
        raise ValueError("No API key provided")


def _is_notion_database(notion_url):
    return "?v=" in notion_url.split("/")[-1]


def use_client(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        orig_client = client = kwargs.pop("client", None)

        if client is None:
            api_key = _load_api_key(kwargs.pop("api_key", None))
            client = Client(auth=api_key)
        out = func(client=client, *args, **kwargs)

        if orig_client is None:
            # Automatically close the client if it was not passed in
            client.close()
        return out

    return wrapper


def query_database(
    database_id: str,
    client: Client,
    start_cursor: Optional[str] = None,
    page_size: int = NOTION_DEFAULT_PAGE_SIZE,
):
    query_dict = {"database_id": database_id, "page_size": page_size}
    if start_cursor is not None:
        query_dict["start_cursor"] = start_cursor
        # For now, Notion API doesn't allow start_cursor='null'

    query_results = client.databases.query(**query_dict)

    assert query_results["object"] == "list"
    return query_results


@use_client
def download(
    notion_url: str,
    nrows: Optional[int] = None,
    *,
    api_key: str = None,
    client: Client = None,
) -> "pd.DataFrame":
    """Download a Notion database as a pandas DataFrame.

    Args:
        notion_url (str): 
            The URL of the Notion database to download from.
        nrows (int, optional): 
            Number of rows of file to read. Useful for reading 
            pieces of large files.
        api_key (str, optional):
            The API key of the Notion integration.
            Defaults to None.
        client (Client, optional):
            The notion client.
            Defaults to None.
    Returns:
        pd.DataFrame: the loaded dataframe.
    """
    assert _is_notion_database(notion_url)
    database_id = get_id(notion_url)

    downloaded_rows = []

    page_size = NOTION_MAX_PAGE_SIZE
    if nrows is not None:
        if nrows <= NOTION_MAX_PAGE_SIZE:
            page_size = nrows    

    query_results = query_database(database_id, client, page_size=page_size)
    downloaded_rows.extend(query_results["results"])

    while query_results["has_more"]:
        if nrows is not None:
            if len(downloaded_rows) >= nrows:
                break
            else:
                page_size = nrows - len(downloaded_rows)
        else:
            page_size = NOTION_MAX_PAGE_SIZE

        query_results = query_database(
            database_id,
            client,
            start_cursor=query_results["next_cursor"],
            page_size=page_size,
        )
        downloaded_rows.extend(query_results["results"])

    properties = PageProperties.from_raw(downloaded_rows)

    retrieve_results = client.databases.retrieve(database_id=database_id)
    schema = DatabaseSchema.from_raw(retrieve_results["properties"])

    df = properties.to_frame()
    df.schema = schema
    return df


def create_database(
    page_id: str, client: Client, schema: DatabaseSchema, title: str = ""
):
    response = client.databases.create(
        parent={"type": "page_id", "page_id": page_id},
        title=[{"type": "text", "text": {"content": title}}],
        properties=schema.query_dict(),
    )
    assert response["object"] == "database"
    return response


def upload_row_to_database(row, database_id, schema, client):

    properties = PageProperty.from_series(row, schema).query_dict()
    client.pages.create(parent={"database_id": database_id}, properties=properties)


def upload_to_database(df, databse_id, schema, client, errors):
    for _, row in df[::NOT_REVERSE_DATAFRAME].iterrows():
        try:
            upload_row_to_database(row, databse_id, schema, client)
        except Exception as e:
            if errors == "strict":
                raise e
            elif errors == "warn":
                warnings.warn(f"Encountered errors {e} while uploading row: {row}")
            elif errors == "ignore":
                continue


def load_database_schema(database_id, client):
    return DatabaseSchema.from_raw(
        client.databases.retrieve(database_id=database_id)["properties"]
    )


@use_client
def upload(
    df: "pd.DataFrame",
    notion_url: str,
    schema: DatabaseSchema = None,
    mode: str = "a",
    title: str = "",
    title_col: str = "",
    errors: str = "strict",
    *,
    api_key: str = None,
    client: Client = None,
):
    """Upload a dataframe to the specified Notion database.

    Args:
        df (pd.DataFrame):
            The dataframe to upload.
        notion_url (str):
            The URL of the Notion page to upload to.
            If it is a notion page, then it will create a new database
            under that page and upload the dataframe to it.
        schema (DatabaseSchema, optional):
            The schema of the Notion database.
            When not set, it will be inferred from (1) the target
            notion database (if it is) then (2) the dataframe itself.
        mode (str, optional):
            (the function is not supported yet.)
            Whether to append to the database or overwrite.
            Defaults to "a".
        title (str, optional):
            The title of the Notion database.
            Defaults to "".
        title_col (str, optional):
            Every Notion database requires a "title" column.
            When the schema is not set, by default it infers the first
            column of uploaded dataframe as the title column. You can
            set this value to specify the title column.
            Defaults to "".
        errors (str, optional):
            Since we upload the dataframe to Notion row by row, you
            can specify how to handle errors during uploading. There
            are several options:
                1. "strict": raise an error when there is one.
                2. "ignore": ignore errors and continue uploading
                    subsequent rows.
                3. "warn": print the error message and continue uploading
            Defaults to "strict".
        api_key (str, optional):
            The API key of the Notion integration.
            Defaults to None.
        client (Client, optional):
            The notion client.
            Defaults to None.
    """
    if schema is None:
        if hasattr(df, "schema"):
            schema = df.schema

    if not _is_notion_database(notion_url):
        if schema is None:
            schema = DatabaseSchema.from_df(df, title_col=title_col)
        database_properties = create_database(get_id(notion_url), client, schema, title)
        databse_id = database_properties["id"]
        notion_url = database_properties["url"]
    else:
        databse_id = get_id(notion_url)
        if schema is None:
            schema = load_database_schema(databse_id, client)

    # At this stage, we should have the appropriate schema
    assert schema is not None

    if not schema.is_df_compatible(df):
        raise ValueError(
            "The dataframe is not compatible with the database schema."
            "The df contains columns that are not in the databse: "
            + f"{[col for col in df.columns if col not in schema.configs.keys()]}"
        )

    if mode not in ("a", "append"):
        raise NotImplementedError
        # TODO: clean the current values in the notion database (if any)

    df = schema.transform(df)
    upload_to_database(df, databse_id, schema, client, errors)

    print(f"Your dataframe has been uploaded to the Notion page: {notion_url} .")
    return notion_url



from typing import List, Dict, Optional, Callable, Tuple
import warnings
import itertools
from dataclasses import dataclass

from pydantic import BaseModel, validator, parse_obj_as
from pandas.api.types import (
    is_datetime64_any_dtype,
    is_numeric_dtype,
    is_object_dtype,
    is_bool_dtype,
    is_categorical_dtype,
    is_list_like,
)

from notion_df.base import (
    SelectOptions,
    NumberFormat,
    RollUpProperty,
    FormulaProperty,
    RelationProperty,
)
from notion_df.utils import (
    flatten_dict,
    SECURE_STR_TRANSFORM,
    SECURE_BOOL_TRANSFORM,
    SECURE_TIME_TRANSFORM,
)


class BasePropertyConfig(BaseModel):
    id: Optional[str]
    type: Optional[str]

    def query_dict(self):
        return flatten_dict(self.dict())

    @validator("type", always=True)
    def automatically_set_type_value(cls, v):
        _type = list(cls.__fields__.keys())[-1]
        if v is None:
            return _type
        else:
            assert _type == v, f"{_type} != {v}"
            return _type


class TitleConfig(BasePropertyConfig):
    title: Dict = {}

    # TODO: Make the validator automatically geneerated
    @validator("title")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The title dict must be empty")
        return v


class RichTextConfig(BasePropertyConfig):
    rich_text: Dict = {}

    @validator("rich_text")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The rich_text dict must be empty")
        return v


class NumberConfig(BasePropertyConfig):
    number: NumberFormat

    # TODO:Add enum based on https://developers.notion.com/reference/create-a-database#number-configuration


class SelectConfig(BasePropertyConfig):
    select: Optional[SelectOptions]


class MultiSelectConfig(BasePropertyConfig):
    multi_select: Optional[SelectOptions]


class DateConfig(BasePropertyConfig):
    date: Dict = {}

    @validator("date")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The date dict must be empty")
        return v


class PeopleConfig(BasePropertyConfig):
    people: Dict = {}

    @validator("people")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The people dict must be empty")
        return v


class FilesConfig(BasePropertyConfig):
    files: Dict = {}

    @validator("files")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The files dict must be empty")
        return v


class CheckboxConfig(BasePropertyConfig):
    checkbox: Dict = {}

    @validator("checkbox")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The checkbox dict must be empty")
        return v


class URLConfig(BasePropertyConfig):
    url: Dict = {}

    @validator("url")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The url dict must be empty")
        return v


class EmailConfig(BasePropertyConfig):
    email: Dict = {}

    @validator("email")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The email dict must be empty")
        return v


class PhoneNumberConfig(BasePropertyConfig):
    phone_number: Dict = {}

    @validator("phone_number")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The phone_number dict must be empty")
        return v


class FormulaConfig(BasePropertyConfig):
    formula: FormulaProperty


class RelationConfig(BasePropertyConfig):
    relation: RelationProperty


class RollupConfig(BasePropertyConfig):
    roll_up: RollUpProperty


class CreatedTimeConfig(BasePropertyConfig):
    created_time: Dict = {}

    @validator("created_time")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The created_time dict must be empty")
        return v


class CreatedByConfig(BasePropertyConfig):
    created_by: Dict = {}

    @validator("created_by")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The created_by dict must be empty")
        return v


class LastEditedTimeConfig(BasePropertyConfig):
    last_edited_time: Dict = {}

    @validator("last_edited_time")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The last_edited_time dict must be empty")
        return v


class LastEditedByConfig(BasePropertyConfig):
    last_edited_by: Dict = {}

    @validator("last_edited_by")
    def title_is_empty_dict(cls, v):
        if v:
            raise ValueError("The last_edited_by dict must be empty")
        return v


def _convert_classname_to_typename(s):
    import re

    s = s.replace("Config", "").replace("URL", "Url")
    return re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()


CONFIGS_MAPPING = {
    _convert_classname_to_typename(_cls.__name__): _cls
    for _cls in BasePropertyConfig.__subclasses__()
}


def parse_single_config(data: Dict) -> BasePropertyConfig:
    return parse_obj_as(CONFIGS_MAPPING[data["type"]], data)


CONFIGS_DF_TRANSFORMER = {
    "title": SECURE_STR_TRANSFORM,
    "rich_text": SECURE_STR_TRANSFORM,
    "number": None,
    "select": SECURE_STR_TRANSFORM,
    "multi_select": lambda lst: [str(ele) for ele in lst] if is_list_like(lst) else str(lst),
    "date": SECURE_TIME_TRANSFORM,
    "checkbox": SECURE_BOOL_TRANSFORM,
    ### TODO: check the following ###
    "people": SECURE_STR_TRANSFORM,
    "files": SECURE_STR_TRANSFORM,
    "url": SECURE_STR_TRANSFORM,
    "email": SECURE_STR_TRANSFORM,
    "phone_number": SECURE_STR_TRANSFORM,
    "formula": SECURE_STR_TRANSFORM,
    "relation": SECURE_STR_TRANSFORM,
    "rollup": SECURE_STR_TRANSFORM,
    "created_time": SECURE_STR_TRANSFORM,
    "created_by": SECURE_STR_TRANSFORM,
    "last_edited_time": SECURE_STR_TRANSFORM,
    "last_edited_by": SECURE_STR_TRANSFORM,
}


def _infer_series_config(column: "pd.Series") -> BasePropertyConfig:
    dtype = column.dtype

    if is_object_dtype(dtype):
        if all(is_list_like(ele) for ele in column):
            all_possible_values = set(
                list(itertools.chain.from_iterable(column.to_list()))
            )
            all_possible_values = [str(ele) for ele in all_possible_values]
            return MultiSelectConfig(
                multi_select=SelectOptions.from_value(all_possible_values),
            )
        else:
            return RichTextConfig()
    if is_numeric_dtype(dtype):
        return NumberConfig(number=NumberFormat(format="number"))
    if is_bool_dtype(dtype):
        return CheckboxConfig()
    if is_categorical_dtype(dtype):
        return SelectConfig(
            select=SelectOptions.from_value([str for cat in dtype.categories]),
        )
    if is_datetime64_any_dtype(dtype):
        return DateConfig()

    return None


@dataclass
class DatabaseSchema:

    configs: Dict[str, BasePropertyConfig]

    @classmethod
    def from_raw(cls, configs: Dict) -> "DatabaseSchema":

        configs = {key: parse_single_config(config) for key, config in configs.items()}
        return cls(configs)

    def __getitem__(self, key: int):
        return self.configs[key]

    def query_dict(self) -> Dict:
        return {key: config.query_dict() for key, config in self.configs.items()}

    @classmethod
    def from_df(
        cls, df: "pd.DataFrame", title_col: Optional[str] = None
    ) -> "DatabaseSchema":
        """Automatically infer the schema from a pandas dataframe"""
        df = df.infer_objects()

        configs = {}
        for col in df.columns:
            config = _infer_series_config(df[col])
            configs[col] = config

        if title_col is not None:
            configs[title_col] = TitleConfig()
        else:
            configs[df.columns[0]] = TitleConfig()
        
        return cls(configs)

    def is_df_compatible(self, df: "pd.DataFrame") -> bool:
        """Validate the dataframe against the schema"""

        if hasattr(df, "schema"):
            if not df.schema == self:
                return False
        else:
            for col in df.columns:
                if col not in self.configs.keys():
                    return False

        # TODO: Add more advanced check on datatypes
        return True

    def transform(self, df: "pd.DataFrame") -> "pd.DataFrame":
        """Transform the df such that the data values are compatible with the schema"""
        df = df.copy()
        for col in df.columns:
            transform = CONFIGS_DF_TRANSFORMER[self[col].type]
            if transform is not None:
                df[col] = df[col].apply(transform)
        return df


from typing import List, Dict, Optional
import re
from enum import Enum
from pydantic import BaseModel, validator
import pandas as pd

from notion_df.utils import is_time_string

### All colors supported in NOTION

class NotionColorEnum(str, Enum):
    Default = "default"
    Gray = "gray"
    Brown = "brown"
    Orange = "orange"
    Yellow = "yellow"
    Green = "green"
    Blue = "blue"
    Purple = "purple"
    Pink = "pink"
    Red = "red"


class NotionExtendedColorEnum(str, Enum):
    Default = "default"
    Gray = "gray"
    Brown = "brown"
    Orange = "orange"
    Yellow = "yellow"
    Green = "green"
    Blue = "blue"
    Purple = "purple"
    Pink = "pink"
    Red = "red"
    GrayBackground = "gray_background"
    BrownBackground = "brown_background"
    OrangeBackground = "orange_background"
    YellowBackground = "yellow_background"
    GreenBackground = "green_background"
    BlueBackground = "blue_background"
    PurpleBackground = "purple_background"
    PinkBackground = "pink_background"
    RedBackground = "red_background"


class RichTextTypeEnum(str, Enum):
    Text = "text"
    Mention = "mention"
    Equation = "equation"


class Annotation(BaseModel):
    bold: bool
    italic: bool
    strikethrough: bool
    underline: bool
    code: bool
    color: NotionExtendedColorEnum


class BaseRichText(BaseModel):
    plain_text: Optional[str]
    # TODO: The Optional[plain_text] is used when creating property values
    href: Optional[str] = None
    annotations: Optional[Annotation] = None
    type: Optional[RichTextTypeEnum]


class Link(BaseModel):
    type: str
    url: str


class Text(BaseModel):
    content: str
    link: Optional[Link]


class RichText(BaseRichText):
    text: Text

    @classmethod
    def from_value(cls, value: str):
        return cls(text=Text(content=value))


class Mention(BaseModel):
    pass  # TODO


class SelectOption(BaseModel):
    id: Optional[str]
    name: str
    color: Optional[NotionColorEnum]

    @classmethod
    def from_value(cls, value: str):
        return cls(name=value)
    
    @validator("name")
    def name_cannot_contain_comma(cls, v):
        if "," in v:
            raise ValueError(f"Invalid option name {v} that contains comma")
        return v


class SelectOptions(BaseModel):
    options: Optional[List[SelectOption]]

    @classmethod
    def from_value(cls, values: List[str]):
        return cls(options=[SelectOption.from_value(value) for value in values])


class NumberFormat(BaseModel):
    format: str


class RollUpProperty(BaseModel):
    relation_property_name: Optional[str]
    relation_property_id: Optional[str]
    rollup_property_name: Optional[str]
    rollup_property_id: Optional[str]
    function: str
    # TODO: Change this to ENUM https://developers.notion.com/reference/create-a-database#rollup-configuration


class FormulaProperty(BaseModel):
    expression: str


class RelationProperty(BaseModel):
    database_id: str
    # TODO: Change this to UUID validation


class Date(BaseModel):
    start: Optional[str] = None
    end: Optional[str] = None
    time_zone: Optional[str] = None

    @validator("start")
    def is_start_ISO8601(cls, v):
        #TODO: Currently it cannot suport time ranges 
        if v is not None:
            if not is_time_string(v):
                raise ValueError(
                    "The data start is not appropriately formatted as an ISO 8601 date string."
                )
        return v

    @validator("end")
    def is_end_ISO8601(cls, v):
        if v is not None:
            if not is_time_string(v):
                raise ValueError(
                    "The data end is not appropriately formatted as an ISO 8601 date string."
                )
        return v

    @classmethod
    def from_value(cls, value: str):
        return cls(start=value)
        # TODO: Now we assume the value has already been formated as strings
        # But we should parse them into appropriate formats.

    @property
    def value(self):
        return pd.to_datetime(self.start)
        # TODO: what should the data structure be if self.end is not None?



from typing import Optional
from notion_df import upload, download


def read_notion(
    notion_url: str,
    nrows: Optional[int] = None,
    api_key: str = None,
) -> "pd.DataFrame":
    """Download a Notion database as a pandas DataFrame.

    Args:
        notion_url (str):
            The URL of the Notion database to download from.
        nrows (int, optional):
            Number of rows of file to read. Useful for reading
            pieces of large files.
        api_key (str, optional):
            The API key of the Notion integration.
            Defaults to None.
        client (Client, optional):
            The notion client.
            Defaults to None.
    Returns:
        pd.DataFrame: the loaded dataframe.
    """
    return download(notion_url, nrows=nrows, api_key=api_key)


def to_notion(
    self,
    notion_url: str,
    schema=None,
    mode: str = "a",
    title: str = "",
    title_col: str = "",
    errors: str = "strict",
    api_key: str = None,
):

    """Upload a dataframe to the specified Notion database.

    Args:
        df (pd.DataFrame):
            The dataframe to upload.
        notion_url (str):
            The URL of the Notion page to upload to.
            If it is a notion page, then it will create a new database
            under that page and upload the dataframe to it.
        schema (DatabaseSchema, optional):
            The schema of the Notion database.
            When not set, it will be inferred from (1) the target
            notion database (if it is) then (2) the dataframe itself.
        mode (str, optional):
            (the function is not supported yet.)
            Whether to append to the database or overwrite.
            Defaults to "a".
        title (str, optional):
            The title of the Notion database.
            Defaults to "".
        title_col (str, optional):
            Every Notion database requires a "title" column.
            When the schema is not set, by default it infers the first
            column of uploaded dataframe as the title column. You can
            set this value to specify the title column.
            Defaults to "".
        errors (str, optional):
            Since we upload the dataframe to Notion row by row, you
            can specify how to handle errors during uploading. There
            are several options:
                1. "strict": raise an error when there is one.
                2. "ignore": ignore errors and continue uploading
                    subsequent rows.
                3. "warn": print the error message and continue uploading
            Defaults to "strict".
        api_key (str, optional):
            The API key of the Notion integration.
            Defaults to None.
    """

    return upload(
        df=self,
        notion_url=notion_url,
        schema=schema,
        mode=mode,
        title=title,
        title_col=title_col,
        errors=errors,
        api_key=api_key,
    )


def pandas():
    import pandas as pd

    pd.read_notion = read_notion
    pd.DataFrame.to_notion = to_notion



from typing import List, Dict, Optional, Union, Any
from datetime import datetime
from dateutil.parser import parse

import pandas as pd
from pandas.api.types import is_array_like, is_datetime64_any_dtype

def flatten_dict(data: Dict):
    """Remove entries in dict whose values are None"""
    if isinstance(data, dict):
        return {
            key: flatten_dict(value)
            for key, value in data.items()
            if value is not None
        }
    elif isinstance(data, list) or isinstance(data, tuple):
        return [flatten_dict(value) for value in data]
    else:
        return data

def is_item_empty(item:Any) -> bool:

    if item is None or item == []:
        return True

    isna = pd.isna(item)
    if is_array_like(isna):
        isna = isna.all()
        # TODO: Rethink it is all or any

    return isna

def is_time_string(s:str) -> bool:
    
    # Ref https://stackoverflow.com/questions/25341945/check-if-string-has-date-any-format
    try: 
        parse(s)
        return True
    except ValueError:
        return False
    

ISO8601_REGEX = r"^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$"
# See https://stackoverflow.com/questions/41129921/validate-an-iso-8601-datetime-string-in-python
ISO8601_STRFTIME_TRANSFORM = lambda ele: ele.strftime('%Y-%m-%dT%H:%M:%SZ')

strtime_transform = lambda ele: parse(ele).strftime('%Y-%m-%dT%H:%M:%SZ') 
datetime_transform = lambda ele: ele.strftime('%Y-%m-%dT%H:%M:%SZ')
def transform_time(s:Any) -> str:
    if not is_item_empty(s):
        if isinstance(s, str):
            return strtime_transform(s)
        elif isinstance(s, datetime):
            return datetime_transform(s)
        elif is_datetime64_any_dtype(s):
            return datetime_transform(s)
    
SECURE_STR_TRANSFORM = lambda ele: str(ele) if not is_item_empty(ele) else ""
SECURE_BOOL_TRANSFORM = lambda ele: bool(ele) if not is_item_empty(ele) else None
SECURE_TIME_TRANSFORM = transform_time


### Referring to https://developers.notion.com/reference/page#property-value-object

from typing import List, Dict, Optional, Union, Any
from dataclasses import dataclass
from collections.abc import Iterable
import numbers

from pydantic import BaseModel, parse_obj_as
import pandas as pd
from pandas.api.types import is_array_like

from notion_df.base import RichText, SelectOption, Date
from notion_df.utils import flatten_dict


class BasePropertyValues(BaseModel):
    id: Optional[str]  # TODO: Rethink whether we can do this
    # The Optional[id] is used when creating property values
    type: Optional[str]

    # TODO: Add abstractmethods for them
    @classmethod
    def from_value(cls, value):
        pass

    @property
    def value(self):
        pass

    def query_dict(self):
        return flatten_dict(self.dict())


class TitleValues(BasePropertyValues):
    title: List[RichText]

    @property
    def value(self) -> Optional[str]:
        return (
            None
            if len(self.title) == 0
            else " ".join([text.plain_text for text in self.title])
        )

    @classmethod
    def from_value(cls, value):
        return cls(title=[RichText.from_value(value)])
        # TODO: Rethink whether we should split input string to multiple elements in the list


class RichTextValues(BasePropertyValues):
    rich_text: List[RichText]

    @property
    def value(self) -> Optional[str]:
        return (
            None
            if len(self.rich_text) == 0
            else " ".join([text.plain_text for text in self.rich_text])
        )

    @classmethod
    def from_value(cls, value: str):
        return cls(rich_text=[RichText.from_value(value)])


class NumberValues(BasePropertyValues):
    number: Optional[Union[float, int]]

    @property
    def value(self) -> str:
        return self.number

    @classmethod
    def from_value(cls, value: Union[float, int]):
        return cls(number=value)


class SelectValues(BasePropertyValues):
    select: Optional[SelectOption]

    @property
    def value(self) -> Optional[str]:
        return self.select.name if self.select else None

    @classmethod
    def from_value(cls, value: str):
        return cls(select=SelectOption.from_value(value))


class MultiSelectValues(BasePropertyValues):
    multi_select: List[SelectOption]

    @property
    def value(self) -> List[str]:
        return [select.name for select in self.multi_select]

    @classmethod
    def from_value(cls, values: Union[List[str], str]):
        if isinstance(values, list):
            return cls(
                multi_select=[SelectOption.from_value(value) for value in values]
            )
        else:
            return cls(multi_select=[SelectOption.from_value(values)])


class DateValues(BasePropertyValues):
    date: Optional[Date]

    @property
    def value(self) -> str:
        return self.date.value if self.date else None

    @classmethod
    def from_value(cls, value: str):
        return cls(date=Date.from_value(value))


class FormulaValues(BasePropertyValues):
    pass


class RelationValues(BasePropertyValues):
    pass


class FileValues(BasePropertyValues):
    pass


class CheckboxValues(BasePropertyValues):
    checkbox: Optional[bool]

    @property
    def value(self) -> Optional[bool]:
        return self.checkbox

    @classmethod
    def from_value(cls, value: bool):
        return cls(checkbox=value)


class URLValues(BasePropertyValues):
    url: Optional[str]

    @property
    def value(self) -> Optional[str]:
        return self.url

    @classmethod
    def from_value(cls, value: str):
        return cls(url=value)


class EmailValues(BasePropertyValues):
    email: Optional[str]

    @property
    def value(self) -> Optional[str]:
        return self.email

    @classmethod
    def from_value(cls, value: str):
        return cls(email=value)


class PhoneNumberValues(BasePropertyValues):
    phone_number: Optional[str]

    @property
    def value(self) -> Optional[str]:
        return self.phone_number

    @classmethod
    def from_value(cls, value: str):
        return cls(phone_number=value)


class CreatedTimeValues(BasePropertyValues):
    created_time: Optional[str]

    @property
    def value(self) -> Optional[str]:
        return self.created_time

    @classmethod
    def from_value(cls, value: str):
        return cls(created_time=value)


class CreatedByValues(BasePropertyValues):
    pass


class LastEditedTimeValues(BasePropertyValues):
    last_edited_time: str

    @property
    def value(self) -> Optional[str]:
        return self.last_edited_time

    @classmethod
    def from_value(cls, value: str):
        return cls(last_edited_time=value)


class LastEditedByValues(BasePropertyValues):
    pass


VALUES_MAPPING = {
    list(_cls.__fields__.keys())[-1]: _cls
    for _cls in BasePropertyValues.__subclasses__()
    if len(_cls.__fields__)
    == 3  # TODO: When all classes have been implemented, we can just remove this check
}


def parse_single_values(data: Dict) -> BasePropertyValues:
    return parse_obj_as(VALUES_MAPPING[data["type"]], data)


def _guess_value_schema(val: Any) -> object:

    if isinstance(val, str):
        return RichTextValues
    elif isinstance(val, numbers.Number):
        return NumberValues
    elif isinstance(val, bool):
        return CheckboxValues
    else:
        raise ValueError(f"Unknown value type: {type(val)}")


def _is_item_empty(item):

    if item is None or item == []:
        return True

    isna = pd.isna(item)
    if is_array_like(isna):
        isna = isna.all()
        # TODO: Rethink it is all or any

    return isna


def parse_value_with_schema(
    idx: int, key: str, value: Any, schema: "DatabaseSchema"
) -> BasePropertyValues:
    # TODO: schema shouldn't be allowed to be empty in the future version
    # schema should be determined at the dataframe level.

    if schema is not None:
        value_func = VALUES_MAPPING[schema[key].type]
    else:
        if idx == 0:
            # TODO: Brutally enforce the first one to be the title, though
            # should be optimized in future versions
            value_func = TitleValues
            value = str(value)
        else:
            value_func = _guess_value_schema(value)

    return value_func.from_value(value)


@dataclass
class PageProperty:
    """This class is used to parse properties of a single Notion Page. 
    
    :: example:
    
        >>> data = \
                {"Description": {"id": "ji%3Dc", "type": "rich_text", "rich_text": []},
                "Created": {"id": "mbOA", "type": "date", "date": None},
                "Title": {"id": "title", "type": "title", "title": []}}
        >>> property = PageProperty.from_raw(data)
    """

    properties: Dict[str, BasePropertyValues]

    @classmethod
    def from_raw(cls, properties: Dict) -> "PageProperty":
        properties = {k: parse_single_values(v) for k, v in properties.items()}
        return cls(properties)

    def __getitem__(self, key):
        return self.properties[key]

    def to_series(self):
        return pd.Series(
            {key: property.value for key, property in self.properties.items()}
        )

    @classmethod
    def from_series(
        cls, series: pd.Series, schema: "DatabaseSchema" = None
    ) -> "PageProperty":
        return cls(
            {
                key: parse_value_with_schema(idx, key, val, schema)
                for idx, (key, val) in enumerate(series.items())
                if not _is_item_empty(val)
            }
        )

    def query_dict(self) -> Dict:
        return {key: property.query_dict() for key, property in self.properties.items()}


@dataclass
class PageProperties:
    """This class is used to parse multiple page properties within a database
    
    :: example:
    
        >>> data = \
                [
                    {
                        "object": "page",
                        "id": "xxxx",
                        "created_time": "2032-01-03T00:00:00.000Z",
                        "properties": {
                            "Description": {"id": "ji%3Dc", "type": "rich_text", "rich_text": []},
                            "Created": {"id": "mbOA", "type": "date", "date": None},
                            "Title": {"id": "title", "type": "title", "title": []}
                        }
                    },
                    {
                        "object": "page",
                        "id": "xxxx",
                        "created_time": "2032-01-03T00:00:01.000Z",
                        "properties": {
                            "Description": {"id": "ji%3Dc", "type": "rich_text", "rich_text": []},
                            "Created": {"id": "mbOA", "type": "date", "date": None},
                            "Title": {"id": "title", "type": "title", "title": []}
                        }
                    }
                ]
        >>> property = PageProperties.from_raw(data)
    """

    page_properties: List[PageProperty]

    @classmethod
    def from_raw(cls, properties: List[Dict]) -> "PageProperties":
        page_properties = [
            PageProperty.from_raw(property["properties"]) for property in properties
        ]
        return cls(page_properties)

    def __getitem__(self, key: int):
        return self.page_properties[key]

    def to_frame(self):
        return pd.DataFrame([property.to_series() for property in self.page_properties])



from setuptools import setup, find_packages
import os


def get_requirements(req_file):
    reqs = []
    with open(req_file, "r") as fp:
        for line in fp.readlines():
            if line.startswith("#") or line.strip() == "":
                continue
            else:
                reqs.append(line.strip())
    return reqs


# A trick from https://github.com/jina-ai/jina/blob/79b302c93b01689e82cf4b52f46522eb7497c404/setup.py#L20
libinfo_py = os.path.join("src", "notion_df", "__init__.py")
libinfo_content = open(libinfo_py, "r", encoding="utf8").readlines()
version_line = [l.strip() for l in libinfo_content if l.startswith("__version__")][0]
exec(version_line)  # gives __version__

setup(
    name="notion-df",
    version=__version__,
    description="Notion-DF: Seamlessly Connecting Notion Database with Pandas DataFrame",
    author="Zejiang Shen",
    author_email="zejiangshen@gmail.com",
    license="MIT",
    url="https://github.com/lolipopshock/notion-df",
    package_dir={"": "src"},
    packages=find_packages("src"),
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
    install_requires=get_requirements("requirements.txt"),
    extras_require={
        "dev": [
            "black==21.12b0",
            "pytest",
        ],
    }
)